@@ -1,13 +1,14 @@
 # This file is part of Tryton.  The COPYRIGHT file at the top level of
 # this repository contains the full copyright notices and license terms.
 from collections import defaultdict
-from trytond.pool import PoolMeta, Pool
-from trytond.model import ModelView, Workflow, fields
-from trytond.pyson import Eval, If, Bool
-from trytond.transaction import Transaction
-from trytond.tools import grouped_slice
+from itertools import groupby
 
+from trytond.model import ModelView, Workflow, fields
 from trytond.modules.product import round_price
+from trytond.pool import Pool, PoolMeta
+from trytond.pyson import Bool, Eval, If
+from trytond.tools import grouped_slice
+from trytond.transaction import Transaction
 
 
 class Invoice(metaclass=PoolMeta):
@@ -21,7 +22,6 @@ class Invoice(metaclass=PoolMeta):
             'invisible': Eval('type') == 'in',
             'readonly': Eval('state', '') != 'draft',
             },
-        depends=['type', 'company', 'state'],
         help="The agent who receives a commission for the invoice.")
 
     @classmethod
@@ -51,25 +51,23 @@ class Invoice(metaclass=PoolMeta):
         return all_commissions
 
     @classmethod
-    @Workflow.transition('paid')
-    def paid(cls, invoices):
+    def set_commissions_date(cls, invoices):
         pool = Pool()
         Date = pool.get('ir.date')
         Commission = pool.get('commission')
-
-        today = Date.today()
-
-        super(Invoice, cls).paid(invoices)
-
         date2commissions = defaultdict(list)
-        for sub_invoices in grouped_slice(invoices):
-            ids = [i.id for i in sub_invoices]
-            for commission in Commission.search([
-                        ('date', '=', None),
-                        ('origin.invoice', 'in', ids, 'account.invoice.line'),
-                        ]):
-                date = commission.origin.invoice.reconciled or today
-                date2commissions[date].append(commission)
+        for company, c_invoices in groupby(invoices, key=lambda i: i.company):
+            with Transaction().set_context(company=company.id):
+                today = Date.today()
+            for sub_invoices in grouped_slice(list(c_invoices)):
+                ids = [i.id for i in sub_invoices]
+                for commission in Commission.search([
+                            ('date', '=', None),
+                            ('origin.invoice', 'in', ids,
+                                'account.invoice.line'),
+                            ]):
+                    date = commission.origin.invoice.reconciled or today
+                    date2commissions[date].append(commission)
         to_write = []
         for date, commissions in date2commissions.items():
             to_write.append(commissions)
@@ -79,6 +77,12 @@ class Invoice(metaclass=PoolMeta):
         if to_write:
             Commission.write(*to_write)
 
+    @classmethod
+    @Workflow.transition('paid')
+    def paid(cls, invoices):
+        super().paid(invoices)
+        cls.set_commissions_date(invoices)
+
     @classmethod
     @ModelView.button
     @Workflow.transition('cancelled')
@@ -86,9 +90,16 @@ class Invoice(metaclass=PoolMeta):
         pool = Pool()
         Commission = pool.get('commission')
 
-        invoices_to_revert_commission = [x for x in invoices if not x.move]
+        invoices_to_revert_commission = []
+        invoices_to_set_date = []
+        for invoice in invoices:
+            if invoice.move:
+                invoices_to_set_date.append(invoice)
+            else:
+                invoices_to_revert_commission.append(invoice)
 
-        super(Invoice, cls).cancel(invoices)
+        super().cancel(invoices)
+        cls.set_commissions_date(invoices_to_set_date)
 
         to_delete = []
         to_save = []
@@ -126,7 +137,8 @@ class InvoiceLine(metaclass=PoolMeta):
             'invisible': If(Bool(Eval('_parent_invoice')),
                 Eval('_parent_invoice', {}).get('type') == 'in',
                 Eval('invoice_type') == 'in'),
-            }, depends=['invoice_type', 'company'],
+            'readonly': Eval('invoice_state') != 'draft',
+            },
         help="The principal who pays a commission for the invoice line.")
     commissions = fields.One2Many('commission', 'origin', 'Commissions',
         readonly=True,
@@ -158,15 +170,17 @@ class InvoiceLine(metaclass=PoolMeta):
         if self.type != 'line':
             return []
 
-        today = Date.today()
+        with Transaction().set_context(company=self.invoice.company.id):
+            today = Date.today()
         commissions = []
         for agent, plan in self.agent_plans_used:
             if not plan:
                 continue
             with Transaction().set_context(date=self.invoice.currency_date):
-                amount = Currency.compute(self.invoice.currency,
+                base_amount = Currency.compute(self.invoice.currency,
                     self.amount, agent.currency, round=False)
-            amount = self._get_commission_amount(amount, plan)
+            base_amount = round_price(base_amount)
+            amount = self._get_commission_amount(base_amount, plan)
             if amount:
                 amount = round_price(amount)
             if not amount:
@@ -181,6 +195,7 @@ class InvoiceLine(metaclass=PoolMeta):
                 commission.date = self.invoice.reconciled or today
             commission.agent = agent
             commission.product = plan.commission_product
+            commission.base_amount = base_amount
             commission.amount = amount
             commissions.append(commission)
         return commissions
